@@ -1,4 +1,4 @@
-# Copyright (C) 2006 Jelmer Vernooij <jelmer@samba.org>
+# Copyright (C) 2006-2007 Jelmer Vernooij <jelmer@samba.org>
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
 from bzrlib.bzrdir import BzrDirFormat, BzrDir
 from bzrlib.errors import NotBranchError, NotLocalUrl, NoRepositoryPresent
 from bzrlib.lockable_files import TransportLock
-from bzrlib.progress import ProgressBar
+from bzrlib.ui import ui_factory
 from bzrlib.transport.local import LocalTransport
 import bzrlib.urlutils as urlutils
 
@@ -27,7 +27,7 @@ import svn.core, svn.repos
 from branch import SvnBranch
 from repository import SvnRepository
 from scheme import BranchingScheme
-from transport import SvnRaTransport, bzr_to_svn_url
+from transport import SvnRaTransport, bzr_to_svn_url, get_svn_ra_transport
 
 
 class SvnRemoteAccess(BzrDir):
@@ -38,16 +38,14 @@ class SvnRemoteAccess(BzrDir):
     """
     def __init__(self, _transport, _format, scheme=None):
         """See BzrDir.__init__()."""
+        _transport = get_svn_ra_transport(_transport)
         super(SvnRemoteAccess, self).__init__(_transport, _format)
 
-        self.transport = None
-        self.svn_root_transport = _transport.get_root()
-
         svn_url = bzr_to_svn_url(self.root_transport.base)
-        root_svn_url = bzr_to_svn_url(self.svn_root_transport.base)
+        self.svn_root_url = _transport.get_repos_root()
 
-        assert svn_url.startswith(root_svn_url)
-        self.branch_path = svn_url[len(root_svn_url):]
+        assert svn_url.startswith(self.svn_root_url)
+        self.branch_path = svn_url[len(self.svn_root_url):]
 
         if scheme is None:
             self.scheme = BranchingScheme.guess_scheme(self.branch_path)
@@ -68,20 +66,20 @@ class SvnRemoteAccess(BzrDir):
     def sprout(self, url, revision_id=None, basis=None, force_new_repo=False):
         """See BzrDir.sprout()."""
         result = BzrDirFormat.get_default_format().initialize(url)
-        repo = self.open_repository()
+        repo = self.find_repository()
         if force_new_repo:
             result_repo = repo.clone(result, revision_id, basis)
         else:
             try:
                 result_repo = result.find_repository()
-                result_repo.fetch(repo, revision_id=revision_id, 
-                                  pb=ProgressBar())
+                result_repo.fetch(repo, revision_id=revision_id)
             except NoRepositoryPresent:
                 result_repo = repo.clone(result, revision_id, basis)
 
         branch = self.open_branch()
-        branch.sprout(result, revision_id)
-        result.create_workingtree()
+        result_branch = branch.sprout(result, revision_id)
+        if result_branch.repository.make_working_trees():
+            result.create_workingtree()
         return result
 
     def open_repository(self):
@@ -89,13 +87,19 @@ class SvnRemoteAccess(BzrDir):
         
         :return: instance of SvnRepository.
         """
-        repos = SvnRepository(self, self.svn_root_transport)
-        repos._format = self._format
-        return repos
+        if self.branch_path == "":
+            return SvnRepository(self, self.root_transport)
+        raise NoRepositoryPresent(self)
 
-    # Subversion has all-in-one, so a repository is always present,
-    # no need to look for it.
-    find_repository = open_repository
+    def find_repository(self):
+        """Open the repository associated with this BzrDir.
+        
+        :return: instance of SvnRepository.
+        """
+        transport = self.root_transport
+        if self.svn_root_url != transport.base:
+            transport = SvnRaTransport(self.svn_root_url)
+        return SvnRepository(self, transport)
 
     def open_workingtree(self):
         """See BzrDir.open_workingtree().
@@ -119,7 +123,9 @@ class SvnRemoteAccess(BzrDir):
         repos = self.open_repository()
         # TODO: Check if there are any revisions in this repository 
         # yet if it is the top-level one
-        return SvnBranch(repos, self.branch_path)
+        branch = SvnBranch(self.root_transport.base, repos, self.branch_path)
+        branch.bzrdir = self
+        return branch
 
     def open_branch(self, unsupported=True):
         """See BzrDir.open_branch()."""
@@ -127,10 +133,8 @@ class SvnRemoteAccess(BzrDir):
         if not self.scheme.is_branch(self.branch_path):
             raise NotBranchError(path=self.root_transport.base)
 
-        repos = self.open_repository()
-
+        repos = self.find_repository()
         branch = SvnBranch(self.root_transport.base, repos, self.branch_path)
- 
         branch.bzrdir = self
         return branch
 
@@ -142,13 +146,20 @@ class SvnFormat(BzrDirFormat):
     def probe_transport(klass, transport):
         format = klass()
 
+        transport = get_svn_ra_transport(transport)
+
         if isinstance(transport, SvnRaTransport):
             return format
 
         raise NotBranchError(path=transport.base)
 
     def _open(self, transport):
-        return SvnRemoteAccess(transport, self)
+        try: 
+            return SvnRemoteAccess(transport, self)
+        except SubversionException, (msg, num):
+            if num == svn.core.SVN_ERR_RA_DAV_REQUEST_FAILED:
+                raise NotBranchError(transport.base)
+            raise
 
     def get_format_string(self):
         return 'Subversion Smart Server'
@@ -165,7 +176,7 @@ class SvnFormat(BzrDirFormat):
 
         local_path = transport._local_base.rstrip("/")
         repos = svn.repos.create(local_path, '', '', None, None)
-        return self.open(SvnRaTransport(svn_to_bzr_url(transport.base)), _found=True)
+        return self.open(SvnRaTransport(transport.base), _found=True)
 
     def is_supported(self):
         """See BzrDir.is_supported()."""
