@@ -18,55 +18,59 @@ import bzrlib
 from bzrlib.branch import BranchCheckResult
 from bzrlib.config import config_dir, ensure_config_dir_exists
 from bzrlib.errors import (BzrError, InvalidRevisionId, NoSuchFile, 
-                           NoSuchRevision, NotBranchError)
+                           NoSuchRevision, NotBranchError, 
+                           UninitializableFormat)
 from bzrlib.graph import Graph
-from bzrlib.inventory import ROOT_ID, Inventory
+from bzrlib.inventory import Inventory
 from bzrlib.lockable_files import LockableFiles, TransportLock
 import bzrlib.osutils as osutils
-from bzrlib.progress import ProgressBar
-from bzrlib.repository import Repository
+from bzrlib.repository import Repository, RepositoryFormat
+from bzrlib.revisiontree import RevisionTree
 from bzrlib.revision import Revision, NULL_REVISION
 from bzrlib.transport import Transport
 from bzrlib.trace import mutter
+from bzrlib.ui import ui_factory
 
 from svn.core import SubversionException, Pool
 import svn.core
 
 import os
 from cStringIO import StringIO
-from bsddb import dbshelve as shelve
+try:
+    import sqlite3
+except ImportError:
+    from pysqlite2 import dbapi2 as sqlite3
 
 import branch
+from branchprops import BranchPropertyList
+import errors
 import logwalker
 from tree import SvnRevisionTree
 
-MAPPING_VERSION = 1
+MAPPING_VERSION = 2
 REVISION_ID_PREFIX = "svn-v%d:" % MAPPING_VERSION
 SVN_PROP_BZR_MERGE = 'bzr:merge'
 SVN_PROP_SVK_MERGE = 'svk:merge'
+SVN_PROP_BZR_FILEIDS = 'bzr:file-ids'
 SVN_PROP_BZR_REVPROP_PREFIX = 'bzr:revprop:'
 SVN_REVPROP_BZR_SIGNATURE = 'bzr:gpg-signature'
 
-_unsafe = "%/-\t "
-def escape_svn_path(id):
-    r = [((c in _unsafe) and ('%%%02x' % ord(c)) or c)
+
+def escape_svn_path(id, unsafe="%/-\t \n"):
+    assert "%" in unsafe
+    r = [((c in unsafe) and ('%%%02x' % ord(c)) or c)
          for c in id]
-    return ''.join(r)
-
-
-def unescape_svn_path(id):
-    ret = ""
-    i = 0
-    while i < len(id): 
-        if id[i] == '%':
-            ret += chr(int(id[i+1:i+3], 16))
-            i+=3
-        else:
-            ret += str(id[i])
-            i+=1
+    ret = ''.join(r)
+    if isinstance(ret, unicode):
+        return ret.encode("utf-8")
     return ret
 
 
+import urllib
+
+unescape_svn_path = urllib.unquote
+
+
 def parse_svn_revision_id(revid):
     """Parse an existing Subversion-based revision id.
 
@@ -109,23 +113,6 @@ def generate_svn_revision_id(uuid, revnum, path):
         return NULL_REVISION
     return "%s%d@%s-%s" % (REVISION_ID_PREFIX, revnum, uuid, escape_svn_path(path.strip("/")))
 
-def parse_revision_id(self, revid):
-    """Parse an existing Subversion-based revision id.
-
-    :param revid: The revision id.
-    :raises: NoSuchRevision
-    :return: Tuple with branch path and revision number.
-    """
-    try:
-        (uuid, branch_path, revnum) = parse_svn_revision_id(revid)
-    except InvalidRevisionId:
-        raise NoSuchRevision(self, revid)
-
-    if uuid != self.uuid:
-        raise NoSuchRevision(self, revid)
-
-    return (branch_path, revnum)
-
 
 def svk_feature_to_revision_id(feature):
     """Create a revision id from a svk feature identifier.
@@ -136,6 +123,7 @@ def svk_feature_to_revision_id(feature):
     (uuid, branch, revnum) = feature.split(":")
     return generate_svn_revision_id(uuid, int(revnum), branch.strip("/"))
 
+
 def revision_id_to_svk_feature(revid):
     """Create a SVK feature identifier from a revision id.
 
@@ -146,6 +134,41 @@ def revision_id_to_svk_feature(revid):
     return "%s:/%s:%d" % (uuid, branch, revnum)
 
 
+def create_cache_dir():
+    ensure_config_dir_exists()
+    cache_dir = os.path.join(config_dir(), 'svn-cache')
+
+    if not os.path.exists(cache_dir):
+        os.mkdir(cache_dir)
+
+        open(os.path.join(cache_dir, "README"), 'w').write(
+"""This directory contains information cached by the bzr-svn plugin.
+
+It is used for performance reasons only and can be removed 
+without losing data.
+
+See http://bazaar-vcs.org/BzrSvn for details.
+""")
+    return cache_dir
+
+
+class SvnRepositoryFormat(RepositoryFormat):
+    rich_root_data = False
+
+    def __init__(self):
+        super(SvnRepositoryFormat, self).__init__()
+        from format import SvnFormat
+        self._matchingbzrdir = SvnFormat()
+
+    def get_format_description(self):
+        return "Subversion Repository"
+
+    def initialize(self, url, shared=False, _internal=False):
+        """Svn repositories cannot be created."""
+        raise UninitializableFormat(self)
+
+cachedbs = {}
+
 class SvnRepository(Repository):
     """
     Provides a simplified interface to a Subversion repository 
@@ -158,14 +181,12 @@ class SvnRepository(Repository):
         assert isinstance(transport, Transport)
 
         control_files = LockableFiles(transport, '', TransportLock)
-        Repository.__init__(self, 'Subversion Smart Server', bzrdir, 
+        Repository.__init__(self, SvnRepositoryFormat(), bzrdir, 
             control_files, None, None, None)
 
-        self.ra = transport.ra
-
-        self.uuid = svn.ra.get_uuid(self.ra)
+        self.transport = transport
+        self.uuid = transport.get_uuid()
         self.base = transport.base
-        self.text_cache = {}
         self.dir_cache = {}
         self.scheme = bzrdir.scheme
         self.pool = Pool()
@@ -173,21 +194,21 @@ class SvnRepository(Repository):
         assert self.base
         assert self.uuid
 
-        mutter("Connected to repository with UUID %s" % self.uuid)
+        cache_file = os.path.join(self.create_cache_dir(), 'cache-v1')
+        if not cachedbs.has_key(cache_file):
+            cachedbs[cache_file] = sqlite3.connect(cache_file)
+        self.cachedb = cachedbs[cache_file]
 
-        mutter('svn latest-revnum')
-        self._latest_revnum = svn.ra.get_latest_revnum(self.ra)
-
-        self._log = logwalker.LogWalker(self.scheme, 
-                                        repos_url=transport.svn_url,
-                                        cache_dir=self.create_cache_dir(), 
+        self._latest_revnum = transport.get_latest_revnum()
+        self._log = logwalker.LogWalker(transport=transport, 
+                                        cache_db=self.cachedb, 
                                         last_revnum=self._latest_revnum)
 
-        self.branchprop_cache = shelve.open(os.path.join(
-                                            self.create_cache_dir(),
-                                            'branchprops'))
+        self.branchprop_list = BranchPropertyList(self._log, self.cachedb)
+        self.fileid_map = SimpleFileIdMap(self, self.cachedb)
 
-        self.fileid_map = SimpleFileIdMap(self._log, self.create_cache_dir())
+    def set_branching_scheme(self, scheme):
+        self.scheme = scheme
 
     def _warn_if_deprecated(self):
         # This class isn't deprecated
@@ -198,21 +219,7 @@ class SvnRepository(Repository):
                            self.base)
 
     def create_cache_dir(self):
-        ensure_config_dir_exists()
-        cache_dir = os.path.join(config_dir(), 'svn-cache')
-
-        if not os.path.exists(cache_dir):
-            os.mkdir(cache_dir)
-
-            open(os.path.join(cache_dir, "README"), 'w').write(
-    """This directory contains information cached by the bzr-svn plugin.
-
-It is used for performance reasons only and can be removed 
-without losing data.
-
-See http://bazaar-vcs.org/BzrSvn for details.
-""")
-
+        cache_dir = create_cache_dir()
         dir = os.path.join(cache_dir, self.uuid)
         if not os.path.exists(dir):
             os.mkdir(dir)
@@ -225,40 +232,17 @@ See http://bazaar-vcs.org/BzrSvn for details.
         assert revision_id != None
         return self.revision_tree(revision_id).inventory
 
-    def get_fileid_map(self, revnum, path, pb=None):
-        return self.fileid_map.get_map(self.uuid, revnum, path, pb)
+    def get_fileid_map(self, revnum, path):
+        return self.fileid_map.get_map(self.uuid, revnum, path,
+                                       self.revision_fileid_renames)
 
-    def transform_fileid_map(self, uuid, revnum, branch, changes, map):
-        return self.fileid_map.apply_changes(uuid, revnum, branch, changes, map)
-
-    def path_to_file_id(self, revnum, path):
-        """Generate a bzr file id from a Subversion file name. 
-        
-        This implementation DOES NOT track renames.
-
-        :param revnum: Revision number.
-        :param path: Absolute path.
-        :return: Tuple with file id and revision id.
-        """
-        assert isinstance(revnum, int)
-        assert isinstance(path, basestring)
-        assert revnum >= 0
-
-        path = path.strip("/")
-
-        (bp, rp) = self.scheme.unprefix(path)
-
-        revid = self.generate_revision_id(revnum, bp)
-
-        map = self.get_fileid_map(revnum, bp)
-
-        try:
-            return map[rp]
-        except KeyError:
-            raise NoSuchFile(path=rp)
+    def transform_fileid_map(self, uuid, revnum, branch, changes, renames):
+        return self.fileid_map.apply_changes(uuid, revnum, branch, changes, 
+                                             renames)
 
     def all_revision_ids(self):
-        raise NotImplementedError(self.all_revision_ids)
+        for (bp, rev) in self.follow_history(self.transport.get_latest_revnum()):
+            yield self.generate_revision_id(rev, bp)
 
     def get_inventory_weave(self):
         raise NotImplementedError(self.get_inventory_weave)
@@ -282,11 +266,11 @@ See http://bazaar-vcs.org/BzrSvn for details.
 
         ancestry = []
 
-        for l in self._get_branch_prop(path, revnum, 
+        for l in self.branchprop_list.get_property(path, revnum, 
                                     SVN_PROP_BZR_MERGE, "").splitlines():
             ancestry.extend(l.split("\n"))
 
-        for (branch, paths, rev) in self._log.follow_history(path, revnum - 1):
+        for (branch, rev) in self.follow_branch(path, revnum - 1):
             ancestry.append(self.generate_revision_id(rev, branch))
 
         ancestry.append(None)
@@ -304,16 +288,13 @@ See http://bazaar-vcs.org/BzrSvn for details.
         except NoSuchRevision:
             return False
 
-        mutter("svn check_path -r%d %s" % (revnum, path))
         try:
-            kind = svn.ra.check_path(self.ra, path.encode('utf8'), revnum)
+            return (svn.core.svn_node_none != self.transport.check_path(path.encode('utf8'), revnum))
         except SubversionException, (_, num):
             if num == svn.core.SVN_ERR_FS_NO_SUCH_REVISION:
                 return False
             raise
 
-        return (kind != svn.core.svn_node_none)
-
     def revision_trees(self, revids):
         for revid in revids:
             yield self.revision_tree(revid)
@@ -324,33 +305,47 @@ See http://bazaar-vcs.org/BzrSvn for details.
 
         if revision_id == NULL_REVISION:
             inventory = Inventory()
+            inventory.revision_id = revision_id
+            return RevisionTree(self, inventory, revision_id)
 
         return SvnRevisionTree(self, revision_id, inventory)
 
-    def revision_parents(self, revision_id, merged_data=None):
-        (path, revnum) = self.parse_revision_id(revision_id)
+    def revision_fileid_renames(self, revid):
+        (path, revnum) = self.parse_revision_id(revid)
+        items = self.branchprop_list.get_property_diff(path, revnum, 
+                                  SVN_PROP_BZR_FILEIDS).splitlines()
+        return dict(map(lambda x: x.split("\t"), items))
 
-        parent_path = None
-        parent_ids = []
-        for (branch, paths, rev) in self._log.follow_history(path, revnum):
+    def _mainline_revision_parent(self, path, revnum):
+        assert isinstance(path, basestring)
+        assert isinstance(revnum, int)
+        for (branch, rev) in self.follow_branch(path, revnum):
             if rev < revnum:
-                parent_revnum = rev
-                parent_path = branch
-                parent_ids = [self.generate_revision_id(rev, branch)]
-                break
+                return self.generate_revision_id(rev, branch)
+        return None
+
+    def revision_parents(self, revision_id, merged_data=None):
+        parent_ids = []
+        (branch, revnum) = self.parse_revision_id(revision_id)
+        mainline_parent = self._mainline_revision_parent(branch, revnum)
+        if mainline_parent is not None:
+            parent_ids.append(mainline_parent)
+            (parent_path, parent_revnum) = self.parse_revision_id(mainline_parent)
+        else:
+            parent_path = None
 
         # if the branch didn't change, bzr:merge can't have changed
-        if not path in self._log.get_revision_info(revnum)[3]:
+        if not self._log.touches_path(branch, revnum):
             return parent_ids
        
         if merged_data is None:
-            new_merge = self._get_branch_prop(path, revnum, 
+            new_merge = self.branchprop_list.get_property(branch, revnum, 
                                            SVN_PROP_BZR_MERGE, "").splitlines()
 
             if len(new_merge) == 0 or parent_path is None:
                 old_merge = ""
             else:
-                old_merge = self._get_branch_prop(parent_path, parent_revnum, 
+                old_merge = self.branchprop_list.get_property(parent_path, parent_revnum, 
                         SVN_PROP_BZR_MERGE, "").splitlines()
 
             assert (len(old_merge) == len(new_merge) or 
@@ -382,24 +377,21 @@ See http://bazaar-vcs.org/BzrSvn for details.
         # Commit SVN revision properties to a Revision object
         rev = Revision(revision_id=revision_id, parent_ids=parent_ids)
 
-        svn_props = self._get_branch_proplist(path, revnum)
+        svn_props = self.branchprop_list.get_properties(path, revnum)
         bzr_props = {}
         for name in svn_props:
             if not name.startswith(SVN_PROP_BZR_REVPROP_PREFIX):
                 continue
 
-            bzr_props[name[len(SVN_PROP_BZR_REVPROP_PREFIX):]] = svn_props[name].decode('utf8')
-
-        (rev.committer, rev.message, date, _) = self._log.get_revision_info(revnum)
+            bzr_props[name[len(SVN_PROP_BZR_REVPROP_PREFIX):]] = svn_props[name]
 
+        (rev.committer, rev.message, date) = self._log.get_revision_info(revnum)
         if rev.committer is None:
             rev.committer = ""
 
         rev.timestamp = 1.0 * svn.core.secs_from_timestr(date, None)
         rev.timezone = None
-
         rev.properties = bzr_props
-
         rev.inventory_sha1 = property(lambda: self.get_inventory_sha1(revision_id))
 
         return rev
@@ -467,8 +459,71 @@ See http://bazaar-vcs.org/BzrSvn for details.
         return bzrlib.xml5.serializer_v5.write_revision_to_string(
             self.get_revision(revision_id))
 
-    def get_revision_sha1(self, revision_id):
-        return osutils.sha_string(self.get_revision_xml(revision_id))
+    def follow_history(self, revnum):
+        while revnum > 0:
+            yielded_paths = []
+            paths = self._log.get_revision_paths(revnum)
+            for p in paths:
+                try:
+                    bp = self.scheme.unprefix(p)[0]
+                    if not bp in yielded_paths:
+                        if not paths.has_key(bp) or paths[bp][0] != 'D':
+                            yield (bp, revnum)
+                        yielded_paths.append(bp)
+                except NotBranchError:
+                    pass
+            revnum-=1
+
+    def follow_branch(self, branch_path, revnum):
+        assert branch_path is not None
+        assert isinstance(revnum, int) and revnum >= 0
+        if not self.scheme.is_branch(branch_path):
+            raise errors.NotSvnBranchPath(branch_path, revnum)
+        branch_path = branch_path.strip("/")
+
+        while revnum > 0:
+            paths = self._log.get_revision_paths(revnum, branch_path)
+            if paths == {}:
+                revnum-=1
+                continue
+            yield (branch_path, revnum)
+            # FIXME: what if one of the parents of branch_path was moved?
+            if (paths.has_key(branch_path) and 
+                paths[branch_path][0] in ('R', 'A')):
+                if paths[branch_path][1] is None:
+                    return
+                if not self.scheme.is_branch(paths[branch_path][1]):
+                    # FIXME: if copyfrom_path is not a branch path, 
+                    # should simulate a reverse "split" of a branch
+                    # for now, just make it look like the branch ended here
+                    return
+                revnum = paths[branch_path][2]
+                branch_path = paths[branch_path][1]
+                continue
+            revnum-=1
+
+    def follow_branch_history(self, branch_path, revnum):
+        assert branch_path is not None
+        if not self.scheme.is_branch(branch_path):
+            raise errors.NotSvnBranchPath(branch_path, revnum)
+
+        for (bp, paths, revnum) in self._log.follow_path(branch_path, revnum):
+            # FIXME: what if one of the parents of branch_path was moved?
+            if (paths.has_key(bp) and 
+                paths[bp][1] is not None and
+                not self.scheme.is_branch(paths[bp][1])):
+                # FIXME: if copyfrom_path is not a branch path, 
+                # should simulate a reverse "split" of a branch
+                # for now, just make it look like the branch ended here
+                for c in self._log.find_children(paths[bp][1], paths[bp][2]):
+                    path = c.replace(paths[bp][1], bp+"/", 1).replace("//", "/")
+                    paths[path] = ('A', None, -1)
+                paths[bp] = ('A', None, -1)
+
+                yield (bp, paths, revnum)
+                return
+                     
+            yield (bp, paths, revnum)
 
     def has_signature_for_revision_id(self, revision_id):
         # TODO: Retrieve from SVN_PROP_BZR_SIGNATURE 
@@ -480,108 +535,6 @@ See http://bazaar-vcs.org/BzrSvn for details.
         # SVN doesn't store GPG signatures
         raise NoSuchRevision(self, revision_id)
 
-    def _cache_get_dir(self, path, revnum):
-        assert path != None
-        path = path.lstrip("/")
-        try:
-           return self.dir_cache[path][revnum]
-        except KeyError:
-            pass
-
-        mutter("svn ls -r %d '%r'" % (revnum, path))
-
-        try:
-            (dirents, _, props) = svn.ra.get_dir(
-                self.ra, path.encode('utf8'), 
-                revnum, self.pool)
-        except SubversionException, (msg, num):
-            if num == svn.core.SVN_ERR_FS_NO_SUCH_REVISION:
-                raise NoSuchRevision(self, revnum)
-            raise
-
-        if not self.dir_cache.has_key(path):
-            self.dir_cache[path] = {}
-
-        self.dir_cache[path][revnum] = (props, dirents)
-
-        return self.dir_cache[path][revnum]
-
-    def _cache_get_file(self, path, revnum):
-        assert path != None
-        path = path.lstrip("/")
-        if self.text_cache.has_key(path) and \
-           self.text_cache[path].has_key(revnum):
-               return self.text_cache[path][revnum]
-
-        stream = StringIO()
-        mutter('svn getfile -r %r %s' % (revnum, path))
-        (realrevnum, props) = svn.ra.get_file(self.ra, path.encode('utf8'), 
-            revnum, stream, self.pool)
-        if not self.text_cache.has_key(path):
-            self.text_cache[path] = {}
-
-        self.text_cache[path][revnum] = (props, stream)
-        return self.text_cache[path][revnum]
-
-    def _get_file_prop(self, path, revnum, name):
-        (props, _) = self._cache_get_file(path, revnum)
-        if props.has_key(name):
-            return props[name]
-        return None
-
-    def _get_branch_proplist(self, path, revnum):
-        i = 0
-        proplist = {}
-        # Search backwards in the cache
-        for i in range(revnum):
-            i = revnum - i
-
-            key = "%s:%d" % (path, i)
-
-            try:
-                proplist = self.branchprop_cache[key]
-            except: 
-                pass
-
-            if self._log.touches_path(path, i):
-                proplist = self._get_dir_proplist(path, i)
-                self.branchprop_cache[key] = proplist
-                break
-
-        return proplist
-
-    def _get_branch_prop(self, path, revnum, name, default=None):
-        props = self._get_branch_proplist(path, revnum)
-        if props.has_key(name):
-            return props[name]
-        return default
-
-    def _get_dir_proplist(self, path, revnum):
-        # find previous ancestor in which properties could've been changed
-        revnum = self._log.find_latest_change(path, revnum)
-
-        (props, _) = self._cache_get_dir(path, revnum)
-        return props
-
-    def _get_dir_prop(self, path, revnum, name, default=None):
-        """Get the Subversion property on a directory.
-
-        Uses the cache.
-        """
-        assert isinstance(revnum, int)
-        assert isinstance(path, basestring)
-
-        props = self._get_dir_proplist(path, revnum)
-        try:
-            return props[name]
-        except KeyError:
-            return default
-
-    def _get_file(self, path, revnum):
-        (_, stream) = self._cache_get_file(path, revnum)
-        stream.seek(0)
-        return stream
-
     def get_revision_graph(self, revision_id):
         if revision_id == NULL_REVISION:
             return {}
@@ -591,7 +544,7 @@ See http://bazaar-vcs.org/BzrSvn for details.
         self._previous = revision_id
         self._ancestry = {}
         
-        for (branch, _, rev) in self._log.follow_history(path, revnum - 1):
+        for (branch, rev) in self.follow_branch(path, revnum - 1):
             revid = self.generate_revision_id(rev, branch)
             self._ancestry[self._previous] = [revid]
             self._previous = revid
@@ -600,6 +553,54 @@ See http://bazaar-vcs.org/BzrSvn for details.
 
         return self._ancestry
 
+    def find_branches(self, revnum=None, pb=None):
+        """Find all branches that were changed in the specified revision number.
+
+        :param revnum: Revision to search for branches.
+        """
+        if revnum is None:
+            revnum = self.transport.get_latest_revnum()
+
+        created_branches = {}
+
+        for i in range(revnum+1):
+            if pb is not None:
+                pb.update("finding branches", i, revnum+1)
+            paths = self._log.get_revision_paths(i)
+            names = paths.keys()
+            names.sort()
+            for p in names:
+                if self.scheme.is_branch(p):
+                    if paths[p][0] in ('R', 'D'):
+                        del created_branches[p]
+                        yield (p, i, False)
+
+                    if paths[p][0] in ('A', 'R'): 
+                        created_branches[p] = i
+                elif self.scheme.is_branch_parent(p):
+                    if paths[p][0] in ('R', 'D'):
+                        k = created_branches.keys()
+                        for c in k:
+                            if c.startswith(p+"/"):
+                                del created_branches[c] 
+                                yield (c, i, False)
+                    if paths[p][0] in ('A', 'R'):
+                        parents = [p]
+                        while parents:
+                            p = parents.pop()
+                            for c in self.transport.get_dir(p, i)[0].keys():
+                                n = p+"/"+c
+                                if self.scheme.is_branch(n):
+                                    created_branches[n] = i
+                                elif self.scheme.is_branch_parent(n):
+                                    parents.append(n)
+
+        for p in created_branches:
+            j = self._log.find_latest_change(p, revnum, recurse=True)
+            if j is None:
+                j = created_branches[p]
+            yield (p, j, True)
+
     def is_shared(self):
         """Return True if this repository is flagged as a shared repository."""
         return True
@@ -607,7 +608,6 @@ See http://bazaar-vcs.org/BzrSvn for details.
     def get_physical_lock_status(self):
         return False
 
-
     def get_commit_builder(self, branch, parents, config, timestamp=None, 
                            timezone=None, committer=None, revprops=None, 
                            revision_id=None):
@@ -631,90 +631,3 @@ See http://bazaar-vcs.org/BzrSvn for details.
         return SvnCommitBuilder(self, branch, parents, config, revprops)
 
 
-class SvnRepositoryRenaming(SvnRepository):
-    """Instance of SvnRepository that tracks renames."""
-
-    def path_to_file_id(self, revnum, path):
-        """Generate a bzr file id from a Subversion file name. """
-        assert isinstance(revnum, int)
-        assert isinstance(path, basestring)
-        assert revnum >= 0
-
-        try:
-            return self.path_map[revnum][path]
-        except:
-            pass
-
-        mutter('creating file id for %r:%d' % (path, revnum))
-
-        (path_branch, filename) = self.scheme.unprefix(path)
-
-        if filename == "":
-            return (ROOT_ID, self.generate_revision_id(revnum, path_branch))
-
-        introduced_revision_id = None
-        last_changed_revid = None
-        continue_revnum = None
-        for (branch, paths, rev) in self._log.follow_history(path_branch, 
-                                                             revnum):
-            if not (continue_revnum is None or continue_revnum == rev):
-                continue
-
-            continue_revnum = None
-
-            expected_path = ("%s/%s" % (branch, filename)).strip("/")
-            parent_changed = False
-            # FIXME: Handle renames of directories
-            for p in paths:
-                otherpath = p[len(branch):].strip("/")
-                if expected_path.startswith(otherpath+"/") and paths[p][0] != 'M':
-                    parent_changed = True
-
-            revid = self.generate_revision_id(rev, branch)
-            if parent_changed:
-                introduced_revision_id = revid
-                break
-
-            if not expected_path in paths:
-                # File changed in this revision
-                continue
-
-            if last_changed_revid is None:
-                mutter('change in %r' % rev)
-                last_changed_revid = revid
-            
-            introduced_revision_id = revid
-
-            # File is being copied from somewhere else
-            if paths[expected_path][1]:
-                copyfrom_path = paths[expected_path][1]
-                copyfrom_rev = paths[expected_path][2]
-                (bp, rp) = self.scheme.unprefix(copyfrom_path)
-
-                # individual non-root dir/file copied from somewhere in 
-                # another branch, don't track a rename
-                if bp != branch:
-                    break
-
-                # check if there is other offspring of that location
-                offspring = self._log.get_offspring(copyfrom_path, copyfrom_rev, rev)
-                # FIXME: Filter out files from offspring that are not in this branch.
-                if list(offspring) != [path]:
-                    break
-
-                filename = rp
-                continue_revnum = copyfrom_rev
-            elif paths[expected_path][0] in ('A', 'R'):
-                break
-
-        assert continue_revnum is None
-        assert last_changed_revid is not None
-
-        if introduced_revision_id is None:
-            raise NoSuchFile(path=filename)
-
-        file_id = "%s-%s" % (introduced_revision_id, escape_svn_path(filename))
-        assert file_id != None
-
-        self.path_map[revnum, path] = (file_id, last_changed_revid)
-        return (file_id, last_changed_revid)
