@@ -25,25 +25,24 @@ import bzrlib.osutils as osutils
 from bzrlib.progress import DummyProgress
 from bzrlib.revision import NULL_REVISION
 from bzrlib.trace import mutter
-from bzrlib.tree import RevisionTree
+from bzrlib.revisiontree import RevisionTree
 
 import os
 import md5
 from cStringIO import StringIO
+import urllib
 
-import svn.core, svn.wc, svn.delta
+import svn.core, svn.wc, svn.delta, svn.ra
 from svn.core import SubversionException, Pool
 
-_global_pool = Pool()
-
-def apply_txdelta_handler(src_stream, target_stream):
+def apply_txdelta_handler(src_stream, target_stream, pool):
     assert hasattr(src_stream, 'read')
     assert hasattr(target_stream, 'write')
     ret = svn.delta.svn_txdelta_apply(
             src_stream, 
             target_stream,
             None,
-            _global_pool)
+            pool)
 
     def wrapper(window):
         svn.delta.invoke_txdelta_window_handler(
@@ -55,38 +54,34 @@ class SvnRevisionTree(RevisionTree):
      def __init__(self, repository, revision_id, inventory=None):
         self._repository = repository
         self._revision_id = revision_id
-        if revision_id == NULL_REVISION:
-            self._inventory = Inventory()
-            self._inventory.revision_id = NULL_REVISION
-        else:
-            (self.branch_path, self.revnum) = repository.parse_revision_id(revision_id)
-            self._inventory = Inventory()
-            self._inventory.revision_id = revision_id
-            self.id_map = repository.get_fileid_map(self.revnum, self.branch_path)
-            self.editor = TreeBuildEditor(self)
-            self.file_data = {}
-
-            editor, baton = svn.delta.make_editor(self.editor)
-
-            root_repos = svn.ra.get_repos_root(repository.ra)
-            mutter('svn checkout -r %r %r' % (self.revnum, self.branch_path))
-            reporter, reporter_baton = svn.ra.do_switch(repository.ra, self.revnum, "", True, os.path.join(root_repos, self.branch_path), editor, baton)
-
-            svn.ra.reporter2_invoke_set_path(reporter, reporter_baton, "", 0, True, None)
-
-            svn.ra.reporter2_invoke_finish_report(reporter, reporter_baton)
+        pool = Pool()
+        (self.branch_path, self.revnum) = repository.parse_revision_id(revision_id)
+        self._inventory = Inventory()
+        self.id_map = repository.get_fileid_map(self.revnum, self.branch_path)
+        self.editor = TreeBuildEditor(self, pool)
+        self.file_data = {}
+        editor, baton = svn.delta.make_editor(self.editor, pool)
+        root_repos = repository.transport.get_repos_root()
+        reporter, reporter_baton = repository.transport.do_switch(
+                self.revnum, "", True, 
+                os.path.join(root_repos, self.branch_path), editor, baton, pool)
+        svn.ra.reporter2_invoke_set_path(reporter, reporter_baton, "", 0, 
+                True, None, pool)
+        svn.ra.reporter2_invoke_finish_report(reporter, reporter_baton, pool)
+        pool.destroy()
 
      def get_file_lines(self, file_id):
         return osutils.split_lines(self.file_data[file_id])
 
 
 class TreeBuildEditor(svn.delta.Editor):
-    def __init__(self, tree):
+    def __init__(self, tree, pool):
         self.tree = tree
         self.repository = tree._repository
         self.last_revnum = {}
         self.dir_revnum = {}
         self.dir_ignores = {}
+        self.pool = pool
 
     def set_target_revision(self, revnum):
         self.revnum = revnum
@@ -94,22 +89,16 @@ class TreeBuildEditor(svn.delta.Editor):
     def open_root(self, revnum, baton):
         return ROOT_ID
 
-    def relpath(self, path):
-        bp, rp = self.tree._repository.scheme.unprefix(path)
-        if bp == self.tree.branch_path:
-            return rp
-        return None
-
     def add_directory(self, path, parent_baton, copyfrom_path, copyfrom_revnum, pool):
+        path = path.decode("utf-8")
         file_id, revision_id = self.tree.id_map[path]
         ie = self.tree._inventory.add_path(path, 'directory', file_id)
-
         ie.revision = revision_id
         return file_id
 
     def change_dir_prop(self, id, name, value, pool):
         from repository import (SVN_PROP_BZR_MERGE, SVN_PROP_SVK_MERGE, 
-                        SVN_PROP_BZR_REVPROP_PREFIX)
+                        SVN_PROP_BZR_REVPROP_PREFIX, SVN_PROP_BZR_FILEIDS)
 
         if name == svn.core.SVN_PROP_ENTRY_COMMITTED_REV:
             self.dir_revnum[id] = int(value)
@@ -119,6 +108,10 @@ class TreeBuildEditor(svn.delta.Editor):
             if id != ROOT_ID:
                 mutter('%r set on non-root dir!' % SVN_PROP_BZR_MERGE)
                 return
+        elif name == SVN_PROP_BZR_FILEIDS:
+            if id != self.tree.id_map[""][0]:
+                mutter('%r set on non-root dir!' % SVN_PROP_BZR_FILEIDS)
+                return
         elif name in (svn.core.SVN_PROP_ENTRY_COMMITTED_DATE,
                       svn.core.SVN_PROP_ENTRY_LAST_AUTHOR,
                       svn.core.SVN_PROP_ENTRY_LOCK_TOKEN,
@@ -127,6 +120,8 @@ class TreeBuildEditor(svn.delta.Editor):
             pass
         elif name.startswith(svn.core.SVN_PROP_WC_PREFIX):
             pass
+        elif name.startswith(SVN_PROP_BZR_REVPROP_PREFIX):
+            pass
         else:
             mutter('unsupported dir property %r' % name)
 
@@ -149,6 +144,7 @@ class TreeBuildEditor(svn.delta.Editor):
             mutter('unsupported file property %r' % name)
 
     def add_file(self, path, parent_id, copyfrom_path, copyfrom_revnum, baton):
+        path = path.decode("utf-8")
         self.is_symlink = False
         self.is_executable = False
         return path
@@ -159,7 +155,6 @@ class TreeBuildEditor(svn.delta.Editor):
 
     def close_file(self, path, checksum):
         file_id, revision_id = self.tree.id_map[path]
-
         if self.is_symlink:
             ie = self.tree._inventory.add_path(path, 'symlink', file_id)
         else:
@@ -181,6 +176,7 @@ class TreeBuildEditor(svn.delta.Editor):
             ie.text_sha1 = None
             ie.text_size = None
             ie.text_id = None
+            ie.executable = False
         else:
             ie.text_sha1 = osutils.sha_string(file_data)
             ie.text_size = len(file_data)
@@ -197,76 +193,89 @@ class TreeBuildEditor(svn.delta.Editor):
 
     def apply_textdelta(self, file_id, base_checksum):
         self.file_stream = StringIO()
-        return apply_txdelta_handler(StringIO(""), self.file_stream)
+        return apply_txdelta_handler(StringIO(""), self.file_stream, self.pool)
 
 
-class SvnInventoryFile(InventoryFile):
-    """Inventory entry that can either be a plain file or a 
-    symbolic link. Avoids fetching data until necessary. """
-    def __init__(self, file_id, name, parent_id, repository, path, revnum, 
-                 has_props):
-        self.repository = repository
-        self.path = path
-        self.has_props = has_props
-        self.revnum = revnum
-        InventoryFile.__init__(self, file_id, name, parent_id)
-
-    def _get_sha1(self):
-        text = self.repository._get_file(self.path, self.revnum).read()
-        return osutils.sha_string(text)
-
-    def _get_executable(self):
-        if not self.has_props:
-            return False
-
-        value = self.repository._get_file_prop(self.path, self.revnum, 
-                    svn.core.SVN_PROP_EXECUTABLE)
-        if value and value == svn.core.SVN_PROP_EXECUTABLE_VALUE:
-            return True
-        return False 
-
-    def _is_special(self):
-        if not self.has_props:
-            return False
-
-        value = self.repository._get_file_prop(self.path, self.revnum, 
-                    svn.core.SVN_PROP_SPECIAL)
-        if value and value == svn.core.SVN_PROP_SPECIAL_VALUE:
-            return True
-        return False 
-
-    def _get_symlink_target(self):
-        if not self._is_special():
-            return None
-        data = self.repository._get_file(self.path, self.revnum).read()
-        if not data.startswith("link "):
-            raise BzrError("Improperly formatted symlink file")
-        return data[len("link "):]
-
-    def _get_kind(self):
-        if self._is_special():
-            return 'symlink'
-        return 'file'
-
-    # FIXME: we need a set function here because of InventoryEntry.__init__
-    def _phony_set(self, data):
-        pass
-   
-    text_sha1 = property(_get_sha1, _phony_set)
-    executable = property(_get_executable, _phony_set)
-    symlink_target = property(_get_symlink_target, _phony_set)
-    kind = property(_get_kind, _phony_set)
-
-
-class SvnBasisTree(SvnRevisionTree):
+class SvnBasisTree(RevisionTree):
     """Optimized version of SvnRevisionTree."""
-    def __init__(self, workingtree, revid):
-        super(SvnBasisTree, self).__init__(workingtree.branch.repository,
-                                           revid)
+    def __init__(self, workingtree):
         self.workingtree = workingtree
+        self._revision_id = workingtree.branch.repository.generate_revision_id(
+                workingtree.base_revnum, workingtree.branch.branch_path)
+        self.id_map = workingtree.branch.repository.get_fileid_map(
+                workingtree.base_revnum, workingtree.branch.branch_path)
+        self._inventory = Inventory()
+        self._repository = workingtree.branch.repository
+
+        def add_file_to_inv(relpath, id, revid, wc):
+            props = svn.wc.get_prop_diffs(self.workingtree.abspath(relpath), wc)
+            if props.has_key(svn.core.SVN_PROP_SPECIAL):
+                ie = self._inventory.add_path(relpath, 'symlink', id)
+                ie.symlink_target = open(self._abspath(relpath)).read()[len("link "):]
+                ie.text_sha1 = None
+                ie.text_size = None
+                ie.text_id = None
+                ie.executable = False
+            else:
+                ie = self._inventory.add_path(relpath, 'file', id)
+                data = osutils.fingerprint_file(open(self._abspath(relpath)))
+                ie.text_sha1 = data['sha1']
+                ie.text_size = data['size']
+                ie.executable = props.has_key(svn.core.SVN_PROP_EXECUTABLE)
+            ie.revision = revid
+            return ie
+
+        def find_ids(entry):
+            relpath = urllib.unquote(entry.url[len(entry.repos):].strip("/"))
+            if entry.schedule in (svn.wc.schedule_normal, 
+                                  svn.wc.schedule_delete, 
+                                  svn.wc.schedule_replace):
+                return self.id_map[workingtree.branch.repository.scheme.unprefix(relpath)[1]]
+            return (None, None)
+
+        def add_dir_to_inv(relpath, wc, parent_id):
+            entries = svn.wc.entries_read(wc, False)
+            entry = entries[""]
+            (id, revid) = find_ids(entry)
+            if id == None:
+                return
+
+            # First handle directory itself
+            ie = self._inventory.add_path(relpath, 'directory', id)
+            ie.revision = revid
+
+            for name in entries:
+                if name == "":
+                    continue
+
+                subrelpath = os.path.join(relpath, name)
+
+                entry = entries[name]
+                assert entry
+                
+                if entry.kind == svn.core.svn_node_dir:
+                    subwc = svn.wc.adm_open3(wc, 
+                            self.workingtree.abspath(subrelpath), 
+                                             False, 0, None)
+                    try:
+                        add_dir_to_inv(subrelpath, subwc, id)
+                    finally:
+                        svn.wc.adm_close(subwc)
+                else:
+                    (subid, subrevid) = find_ids(entry)
+                    if subid is not None:
+                        add_file_to_inv(subrelpath, subid, subrevid, wc)
+
+        wc = workingtree._get_wc() 
+        try:
+            add_dir_to_inv("", wc, None)
+        finally:
+            svn.wc.adm_close(wc)
+
+    def _abspath(self, relpath):
+        return svn.wc.get_pristine_copy_path(self.workingtree.abspath(relpath))
 
     def get_file_lines(self, file_id):
-        path = self.id2path(file_id)
-        base_copy = svn.wc.get_pristine_copy_path(self.workingtree.abspath(path))
+        base_copy = self._abspath(self.id2path(file_id))
         return osutils.split_lines(open(base_copy).read())
 
